@@ -22,10 +22,11 @@ from string import capwords
 from urllib import urlencode
 from cookielib import CookieJar
 from urlparse import urlparse, urljoin, urlunparse
-from types import (StringTypes, TupleType, DictType, NoneType, ListType)
+from types import (StringTypes, TupleType, DictType, NoneType,
+                   ListType, FunctionType)
 
 import pycurl
-
+from .auth import AuthManager, BasicAuth, DigestAuth
 from .exceptions import (HTTPError, InvalidMethod, CurlError, InterfaceError)
 from .utils import (decode_gzip, CaseInsensitiveDict, to_cookiejar,
                    morsel_to_cookie, data_wrapper, make_curl_post_files, utf8, to_unicode)
@@ -38,7 +39,6 @@ except ImportError:
 
 try:
     import signal
-    from signal import SIGPIPE, SIG_IGN
     signal.signal(signal.SIGPIPE, signal.SIG_IGN)
 except ImportError:
     pass
@@ -126,8 +126,7 @@ class Request(object):
 
     >>> request = Request("GET", "http://google.com")
     >>> print(repr(request))
-    ... <Request: GET [ http://google.com ]>
-    ...
+    <Request: GET [ http://google.com ]>
     >>> request.send()
     >>> response = requests.response
     """
@@ -137,8 +136,8 @@ class Request(object):
     def __init__(self, method, url, params=None, data=None, headers=None, cookies=None,
                  files=None, timeout=None, connection_timeout=None, allow_redirects=False,
                  max_redirects=5, proxies=None, auth=None, network_interface=None, use_gzip=None,
-                 validate_cert=False, ca_certs=None, debug_curl=False, user_agent=None, ip_v6=False):
-        r"""A single HTTP / HTTPS request
+                 validate_cert=False, ca_certs=None, cert=None, debug=False, user_agent=None, ip_v6=False, **kwargs):
+        """A single HTTP / HTTPS request
 
         Arguments:
         - `url`: (string) resource url
@@ -164,20 +163,23 @@ class Request(object):
                (('http', ('127.0.0.1', 9050)),
                 ('socks', ('127.0.0.1', 9050, ('username', 'password')))
         - `auth`: (dict, tuple or list) for resource base auth
-        - `network_interface`: (str) use given interface for request
+        - `network_interface`: (str) Pepform an operation using a specified interface.
+           You can enter interface name, IP address or host name.
         - `use_gzip`: (bool) accept gzipped data
         - `validate_cert`: (bool)
-        - `ca_certs`:
-        - `debug_curl`: (bool) use for `pycurl.DEBUGFUNCTION`
+        - `ca_certs`: tells curl to use the specified certificate file to verify the peer.
+        - `cert`: (string) tells curl to use the specified certificate file
+           when getting a file with HTTPS.
+        - `debug`: (bool) use for `pycurl.DEBUGFUNCTION`
         - `user_agent`: (string) user agent
         - `ip_v6`: (bool) use ipv6 protocol
         """
         self._url = url
         if not method or not isinstance(method, StringTypes):
-            raise RuntimeError("method argument must be string")
+            raise InterfaceError("method argument must be string")
 
         if method.upper() not in self.SUPPORTED_METHODS:
-            raise InvalidMethod("cURL do not support %s method" % self._method)
+            raise InvalidMethod("cURL do not support %s method" % method.upper())
 
         self._method = method.upper()
 
@@ -198,7 +200,9 @@ class Request(object):
         else:
             self._data = data_wrapper(data)
 
-        if isinstance(cookies, (TupleType, DictType, CookieJar)):
+        if isinstance(cookies, CookieJar):
+            self._cookies = cookies
+        elif isinstance(cookies, (TupleType, DictType)):
             self._cookies = to_cookiejar(cookies)
         else:
             self._cookies = None
@@ -210,10 +214,10 @@ class Request(object):
 
         self._network_interface = network_interface
 
-        if isinstance(auth, DictType):
+        if isinstance(auth, AuthManager):
             self._auth = auth
         elif isinstance(auth, TupleType):
-            self._auth = dict(username=auth[0], password=auth[1])
+            self._auth = BasicAuth(*auth)
         elif auth is None:
             self._auth = None
         else:
@@ -227,10 +231,13 @@ class Request(object):
         self._connection_timeout = connection_timeout
 
         self._use_gzip = use_gzip
+
+        # Certificates
         self._validate_cert = validate_cert
         self._ca_certs = ca_certs
+        self._cert = None
         self._start_time = time.time()
-        self._debug_curl = debug_curl
+        self._debug_curl = debug
         self._ip_v6 = ip_v6
 
         self.response = None
@@ -243,7 +250,7 @@ class Request(object):
     def user_agent(self):
         # from curl_requests import get_version
         if self._user_agent is None:
-            self._user_agent = "Mozilla/5.0 (compatible; curl_requests)"
+            self._user_agent = "Mozilla/5.0 (compatible; human_curl; +http://h.wrttn.me/human_curl)"
         return self._user_agent
 
     def _build_url(self):
@@ -298,8 +305,10 @@ class Request(object):
         except pycurl.error, e:
             raise CurlError(e[0], e[1])
         else:
-            self.response = Response(url=self._url, curl_opener=opener, body_output=body_output,
-                                     headers_output=headers_output, request=self, cookies_jar=self._cookies)
+            self.response = Response(url=self._url, curl_opener=opener,
+                                     body_output=body_output,
+                                     headers_output=headers_output, request=self,
+                                     cookies=self._cookies)
 
         return self.response
 
@@ -355,9 +364,15 @@ class Request(object):
             opener.setopt(pycurl.CONNECTTIMEOUT, self._connection_timeout)
 
         # Setup debug output write function
-        if self._debug_curl is True:
+        if isinstance(self._debug_curl, FunctionType):
+            logger.debug("Setup %s as debug function" % self._debug_curl.__name__)
+            opener.setopt(pycurl.VERBOSE, 1)
+            opener.setopt(pycurl.DEBUGFUNCTION, self._debug_curl)
+        elif self._debug_curl is True:
             opener.setopt(pycurl.VERBOSE, 1)
             opener.setopt(pycurl.DEBUGFUNCTION, _debug_curl)
+        else:
+            opener.setopt(pycurl.VERBOSE, 0)
 
         # Send allow gzip encoding header
         if self._use_gzip is not None:
@@ -387,6 +402,8 @@ class Request(object):
             if proxy_auth:
                 if len(proxy_auth) == 2:
                     opener.setopt(pycurl.PROXYUSERPWD, "%s:%s" % proxy_auth)
+                else:
+                    raise InterfaceError("Proxy auth data must be tuple")
 
         logger.debug("Setup user agent %s" % self.user_agent)
         opener.setopt(pycurl.USERAGENT, self.user_agent)
@@ -405,6 +422,15 @@ class Request(object):
             if file_exists(self._ca_certs):
                 opener.setopt(pycurl.CAINFO, self._ca_certs)
 
+        ## (HTTPS) Tells curl to use the specified certificate file when getting a
+        ## file with HTTPS. The certificate must be in PEM format.
+        ## If the optional password isn't specified, it will be queried for on the terminal.
+        ## Note that this certificate is the private key and the private certificate concatenated!
+        ## If this option is used several times, the last one will be used.
+        if self._cert:
+            opener.setopt(pycurl.SSLCERT, self._cert)
+
+
         if self._ip_v6:
             opener.setopt(pycurl.IPRESOLVE, pycurl.IPRESOLVE_WHATEVER)
         else:
@@ -425,7 +451,8 @@ class Request(object):
                 name = urllib.quote_plus(name)
                 value = urllib.quote_plus(value)
                 chunks.append('%s=%s;' % (name, value))
-            opener.setopt(pycurl.COOKIE, ''.join(chunks))
+            if chunks:
+                opener.setopt(pycurl.COOKIE, ''.join(chunks))
         else:
             # set empty cookie to activate cURL cookies
             opener.setopt(pycurl.COOKIELIST, '')
@@ -471,12 +498,9 @@ class Request(object):
 
                     opener.setopt(pycurl.IOCTLFUNCTION, ioctl)
                     if self._method == "PUT":
-                        logger.debug("Setup INFILESIZE")
                         opener.setopt(pycurl.PUT, True)
                         opener.setopt(pycurl.INFILESIZE, len(self._data))
                     else:
-                        logger.debug("Setup postfieldsize")
-                        #opener.setopt(pycurl.INFILESIZE, len(self._data))
                         opener.setopt(pycurl.POST, True)
                         opener.setopt(pycurl.POSTFIELDSIZE, len(self._data))
                 elif isinstance(self._data, (TupleType, ListType, DictType)):
@@ -486,12 +510,8 @@ class Request(object):
                     # use postfields to send vars as application/x-www-form-urlencoded
                     # opener.setopt(pycurl.POSTFIELDS, encoded_data)
 
-        if self._auth is not None:
-            logger.debug("Use %s for auth" % self._auth)
-            if self._auth['username'] and self._auth['password']:
-                opener.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_BASIC)
-                opener.setopt(pycurl.USERPWD, "%s:%s" % (self._auth['username'],
-                                                         self._auth['password']))
+        if isinstance(self._auth, AuthManager):
+            self._auth.setup(opener)
         else:
             opener.unsetopt(pycurl.USERPWD)
 
@@ -503,7 +523,7 @@ class Response(object):
     """
 
     def __init__(self, url, curl_opener, body_output, headers_output,
-                 request=None, cookies_jar=None):
+                 request=None, cookies=None):
         """
         Arguments:
         - `url`: resource url
@@ -526,7 +546,12 @@ class Response(object):
 
         # Cookies dictionary
         self._cookies = None
-        self._cookies_jar = cookies_jar
+        if isinstance(cookies, CookieJar):
+            self._cookies_jar = cookies
+        elif isinstance(cookies, (TupleType, DictType)):
+            self._cookies_jar = to_cookiejar(cookies)
+        else:
+            self._cookies_jar = None
 
         # Seconds from request start to finish
         self.request_time = None
@@ -588,7 +613,7 @@ class Response(object):
         return self._status_code
 
     @property
-    def cookiejar(self):
+    def cookiesjar(self):
         """Returns cookie jar object
         """
         if not self._cookies_jar:
@@ -602,7 +627,8 @@ class Response(object):
         """
         import zlib
         if not self._content:
-            if 'gzip' in self.headers.get('Content-Encoding', ''):
+            if 'gzip' in self.headers.get('Content-Encoding', '') and \
+                   'zlib' not in pycurl.version:
                 try:
                     self._content = decode_gzip(self._body_otput.getvalue())
                 except zlib.error:
@@ -636,6 +662,8 @@ class Response(object):
                         if not value.startswith("http"):
                             value = urljoin(self.url, value)
                         self._history.append(value)
+                    if value[:1] == value[-1:] == '"':
+                        value = value[1:-1] # strip "
                     block_headers.append((field, value.strip()))
                 elif header.startswith("HTTP"):
                     # extract version, code, message from first header
@@ -673,6 +701,7 @@ class Response(object):
                     cookie = SimpleCookie()
                     cookie.load(value)
                     cookies.extend(cookie.values())
+
                     # update cookie jar
                     for morsel in cookie.values():
                         if isinstance(self._cookies_jar, CookieJar):
@@ -709,6 +738,11 @@ class Response(object):
 
 
 def _debug_curl(debug_type, debug_msg):
+    """Handle debug messages
+
+    - `debug_type`: (int) debug output code
+    - `debug_msg`: (str) debug message
+    """
     debug_types = ('I', '<', '>', '<', '>')
     if debug_type == 0:
         logger.debug('%s', debug_msg.strip())
